import 'isolate_bloc.dart';
import 'isolate_bloc_connector.dart';
import 'isolate_bloc_wrapper.dart';
import 'isolate_manager.dart';
import 'isolate_service_event_listener.dart';
import 'service_event_listener.dart';
import 'service_events.dart';

/// Signature for initialization function which would be runned in [Isolate] to
/// initialize your blocs and repository.
/// Initializer must be a global or static function.
typedef Initializer = Function();

/// Maintain [IsolateBlocWrapper]s
class BlocManager {
  static BlocManager instance;

  final ServiceEventListener _serviceEventListener;
  final IsolateManager _isolateManager;
  final Map<Type, Object> _initialStates;
  final _freeWrappers = <Type, IsolateBlocWrapper>{};
  final _wrappers = <String, IsolateBlocWrapper>{};

  BlocManager._(
      this._initialStates, this._serviceEventListener, this._isolateManager);

  /// Initialize [Isolate], ServiceEventListener in both Isolates and run [Initializer].
  /// If already initialized kill previous [Isolate] and reinitialize everything.
  static Future<void> initialize(Initializer initializer) async {
    assert(
      '$initializer'.contains(' static'),
      '$Initializer must be a static or global function',
    );

    if (instance != null) {
      instance.dispose();
    }

    final isolateManager =
        await IsolateManager.createIsolate(_isolatedBlocRunner, initializer);

    final serviceEventsListener = ServiceEventListener(
        isolateManager.messenger.sendMessage,
        Stream.castFrom<Object, ServiceEvent>(isolateManager
            .messenger.receiveMessages
            .where((event) => event is ServiceEvent)));
    var initialStates = await serviceEventsListener.getInitialStates();
    instance = BlocManager._(
      initialStates,
      serviceEventsListener,
      isolateManager,
    );
  }

  /// Start creating bloc and return its initialState
  IsolateBlocWrapper<State> createBloc<T extends IsolateBloc, State>() {
    assert(T != Object && T != dynamic, "BlocT [Bloc Type] must not be null");

    final initialState = _initialStates[T];
    final messageReceiver = _serviceEventListener.sendEvent;
    final onBlocClose =
        (uuid) => _serviceEventListener.sendEvent(CloseIsolateBlocEvent(uuid));
    var blocWrapper =
        IsolateBlocWrapper<State>(initialState, messageReceiver, onBlocClose);
    _freeWrappers[T] = blocWrapper;
    _serviceEventListener.sendEvent(CreateIsolateBlocEvent(T));
    return blocWrapper;
  }

  void bindFreeWrapper(Type blocType, String id) {
    assert(_freeWrappers.containsKey(blocType),
        'No free bloc wrapper for $blocType');
    _wrappers[id] = _freeWrappers.remove(blocType)..setUuid(id);
  }

  IsolateBlocWrapper getWrapper(String id) {
    return _wrappers[id];
  }

  static void _isolatedBlocRunner(
    IsolateMessenger messenger,
    Initializer userInitializer,
  ) async {
    await userInitializer();
    IsolateBlocConnector.instance.initializeCompleted();
    messenger.sendMessage(
        IsolateBlocsInitialized(IsolateBlocConnector.instance.initialStates));
    IsolateServiceEventListener(
      messenger.sendMessage,
      Stream.castFrom<Object, ServiceEvent>(
          messenger.receiveMessages.where((event) => event is ServiceEvent)),
    );
  }

  void dispose() {
    _isolateManager.isolate.kill();
    _serviceEventListener.dispose();
  }
}
