import 'package:flutter/foundation.dart';
import 'bloc_manager.dart';
import 'isolate_bloc_wrapper.dart';

import 'isolate_bloc.dart';
import 'isolate_bloc_connector.dart';

void register<Event, State>({@required IsolateBlocCreator create}) =>
    IsolateBlocConnector.instance.register<Event, State>(create);

IsolateBlocWrapper<State>
    createBloc<BlocT extends IsolateBloc<Object, State>, State>() {
  var blocManager = BlocManager.instance;
  assert(
      blocManager != null,
      "$BlocManager must not be null. "
      "Call `await $initialize()` and call this function");
  return blocManager.createBloc<BlocT, State>();
}

/// Initialize [Isolate], ServiceEventListener in both Isolates and run [Initializer].
/// If already initialized and [reCreate] is true kill previous [Isolate] and reinitialize everything.
Future<void> initialize(Initializer userInitializer,
    {bool reCreate = false}) async {
  if (!reCreate) {
    assert(
        BlocManager.instance == null,
        "You can initialize only once. "
        "Call `initialize(..., reCreate: true)` if you want to reinitialize.");
  }
  return BlocManager.initialize(userInitializer);
}

/// Signature for [IsolateBloc] injection.
typedef BlocInjector<T extends IsolateBloc> = Future<T>
    Function<T extends IsolateBloc>();

/// Use this function to get IsolateBloc in Isolate.
/// To get bloc in UI Isolate use IsolateBlocProvider.
/// This function works this way: firstly it is wait for user's initialization function
/// secondly it is looks for created bloc with type BlocA. If it is finds any, so it
/// returns this bloc. Else it is creates a new bloc and add to the pull of free blocs.
/// So when UI will call `create<BlocA, BlocAState>()`, it will not create a new bloc but
/// return free bloc from pull
Future<T> getBloc<T extends IsolateBloc>() async {
  assert(
    IsolateBlocConnector.instance != null,
    "$IsolateBlocConnector instance is null. "
    "Make sure that you call this function from $Initializer.",
  );
  return IsolateBlocConnector.instance.getBloc<T>();
}

/// Signature for [IsolateBlocWrapper] injection.
typedef WrapperInjector<Bloc extends IsolateBloc<Object, State>, State>
    = IsolateBlocWrapper<State> Function<Bloc extends IsolateBloc, State>();

/// Use this function if you need to get [IsolateBloc] sync and you need only
/// `add()` and `listen()` functionality.
/// This function works the same way as [getBloc] and have same effects.
IsolateBlocWrapper<State> getBlocWrapper<Bloc extends IsolateBloc, State>() {
  assert(
    IsolateBlocConnector.instance != null,
    "$IsolateBlocConnector instance is null. "
    "Make sure that you call this function from $Initializer.",
  );
  return IsolateBlocConnector.instance.getBlocWrapper<Bloc, State>();
}
