import 'dart:async';

import 'package:flutter/foundation.dart';
import 'isolate_bloc.dart';
import 'service_events.dart';

import 'transition.dart';

/// Signature for event receiver function which takes an [IsolateBlocTransitionEvent]
/// and send this event to the [IsolateBloc]
typedef EventReceiver = void Function(IsolateBlocTransitionEvent<Object> event);

/// Signature for function which takes [IsolateBloc]'s uuid and close this it
typedef BlocClosure = void Function(String uuid);

/// {@template bloc}
/// This is implementation of
/// {@endtemplate}
class IsolateBlocWrapper<State> extends Stream<State> implements Sink<Object> {
  final _eventController = StreamController<Object>.broadcast();
  final _stateController = StreamController<State>.broadcast();

  /// Id of IsolateBloc. It's needed to find bloc in isolate.
  String _originBlocUuid;

  State _state;
  bool _initStateProvided;
  final _unsentEvents = <Object>[];
  final BlocClosure _onBlocClose;
  StreamSubscription<Transition<Object, State>> _stateTransitionSubscription;
  StreamSubscription<Object> _eventReceiverSubscription;

  /// Returns stream with `event`
  Stream<Object> get eventStream => _eventController.stream;

  /// Callback which receive events and send them to the IsolateBloc
  final EventReceiver _eventReceiver;

  /// Returns the current [state] of the [bloc].
  State get state => _state;

  /// Returns whether the `Stream<State>` is a broadcast stream.
  @override
  bool get isBroadcast => _stateController.stream.isBroadcast;

  /// {@macro bloc}
  IsolateBlocWrapper(
    this._state,
    this._eventReceiver,
    this._onBlocClose,
  ) : _initStateProvided = true {
    _bindEventsListener();
  }

  IsolateBlocWrapper.noInitState(
    this._eventReceiver,
    this._onBlocClose,
  )   : _state = null,
        _initStateProvided = false {
    _bindEventsListener();
  }

  /// Adds a subscription to the `Stream<State>`.
  /// Returns a [StreamSubscription] which handles events from
  /// the `Stream<State>` using the provided [onData], [onError] and [onDone]
  /// handlers.
  @override
  StreamSubscription<State> listen(
    void Function(State) onData, {
    Function onError,
    void Function() onDone,
    bool cancelOnError,
  }) {
    return _prepareStateStream.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  Stream<State> get _prepareStateStream async* {
    if (_initStateProvided) {
      yield state;
    }
    yield* _stateController.stream;
  }

  /// Notifies the [bloc] of a new [event] which triggers onEvent in [IsolateBloc].
  @override
  void add(Object event) {
    _eventController.add(event);
  }

  /// Closes the `event` stream and request to close [IsolateBloc]
  @override
  @mustCallSuper
  Future<void> close() async {
    _onBlocClose(_originBlocUuid);
    await _eventController.close();
    await _stateController.close();
    await _stateTransitionSubscription?.cancel();
    await _eventReceiverSubscription?.cancel();
  }

  /// Connect this wrapper to the origin IsolateBloc and start listening for state.
  void setUuid(String uuid) {
    assert(uuid != null);
    _originBlocUuid = uuid;
    while (_unsentEvents.isNotEmpty) {
      _eventReceiver(IsolateBlocTransitionEvent<Object>(
        _originBlocUuid,
        _unsentEvents.removeAt(0),
      ));
    }
  }

  /// Receive IsolateBloc's state and map them
  void stateReceiver(State nextState) {
    _initStateProvided = true;
    _stateController.add(nextState);
  }

  /// Start listening for new `events`
  void _bindEventsListener() {
    _eventReceiverSubscription = eventStream.listen((event) {
      if (_originBlocUuid != null) {
        _eventReceiver(
            IsolateBlocTransitionEvent<Object>(_originBlocUuid, event));
      } else {
        _unsentEvents.add(event);
      }
    });
  }
}
