import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:nested/nested.dart';
import '../common/isolate_bloc.dart';
import '../common/isolate_bloc_wrapper.dart';
import 'isolate_bloc_provider.dart';

/// Mixin which allows `MultiBlocListener` to infer the types
/// of multiple [IsolateBlocListener]s.
mixin BlocListenerSingleChildWidget on SingleChildWidget {}

/// Signature for the `listener` function which takes the `BuildContext` along
/// with the `state` and is responsible for executing in response to
/// `state` changes.
typedef BlocWidgetListener<S> = void Function(BuildContext context, S state);

/// Signature for the `listenWhen` function which takes the previous `state`
/// and the current `state` and is responsible for returning a [bool] which
/// determines whether or not to call [BlocWidgetListener] of [IsolateBlocListener]
/// with the current `state`.
typedef BlocListenerCondition<S> = bool Function(S previous, S current);

/// {@template bloc_listener}
/// Takes a [BlocWidgetListener] and an optional [isolateBloc] and invokes
/// the [listener] in response to `state` changes in the [isolateBloc].
/// It should be used for functionality that needs to occur only in response to
/// a `state` change such as navigation, showing a `SnackBar`, showing
/// a `Dialog`, etc...
/// The [listener] is guaranteed to only be called once for each `state` change
/// unlike the `builder` in `BlocBuilder`.
///
/// If the [isolateBloc] parameter is omitted, [IsolateBlocListener] will automatically
/// perform a lookup using [BlocProvider] and the current `BuildContext`.
///
/// ```dart
/// BlocListener<BlocA, BlocAState>(
///   listener: (context, state) {
///     // do stuff here based on BlocA's state
///   },
///   child: Container(),
/// )
/// ```
/// Only specify the [isolateBloc] if you wish to provide a [isolateBloc] that is otherwise
/// not accessible via [BlocProvider] and the current `BuildContext`.
///
/// ```dart
/// BlocListener<BlocA, BlocAState>(
///   bloc: blocA,
///   listener: (context, state) {
///     // do stuff here based on BlocA's state
///   },
///   child: Container(),
/// )
/// ```
/// {@endtemplate}
/// {@template bloc_listener_listen_when}
/// An optional [listenWhen] can be implemented for more granular control
/// over when [listener] is called.
/// [listenWhen] will be invoked on each [isolateBloc] `state` change.
/// [listenWhen] takes the previous `state` and current `state` and must
/// return a [bool] which determines whether or not the [listener] function
/// will be invoked.
/// The previous `state` will be initialized to the `state` of the [isolateBloc]
/// when the [IsolateBlocListener] is initialized.
/// [listenWhen] is optional and if omitted, it will default to `true`.
///
/// ```dart
/// BlocListener<BlocA, BlocAState>(
///   listenWhen: (previous, current) {
///     // return true/false to determine whether or not
///     // to invoke listener with state
///   },
///   listener: (context, state) {
///     // do stuff here based on BlocA's state
///   }
///   child: Container(),
/// )
/// ```
/// {@endtemplate}
class IsolateBlocListener<C extends IsolateBloc<Object, S>, S>
    extends IsolateBlocListenerBase<C, S> with BlocListenerSingleChildWidget {
  /// {@macro bloc_listener}
  const IsolateBlocListener({
    Key key,
    @required BlocWidgetListener<S> listener,
    IsolateBlocWrapper<S> isolateBloc,
    BlocListenerCondition<S> listenWhen,
    this.child,
  })  : assert(listener != null),
        super(
          key: key,
          child: child,
          listener: listener,
          isolateBloc: isolateBloc,
          listenWhen: listenWhen,
        );

  /// The widget which will be rendered as a descendant of the [IsolateBlocListener].
  @override
  // ignore: overridden_fields
  final Widget child;
}

/// {@template bloc_listener_base}
/// Base class for widgets that listen to state changes in a specified [isolateBloc].
///
/// A [IsolateBlocListenerBase] is stateful and maintains the state subscription.
/// The type of the state and what happens with each state change
/// is defined by sub-classes.
/// {@endtemplate}
abstract class IsolateBlocListenerBase<C extends IsolateBloc<Object, S>, S>
    extends SingleChildStatefulWidget {
  /// {@macro bloc_listener_base}
  const IsolateBlocListenerBase({
    Key key,
    this.listener,
    this.isolateBloc,
    this.child,
    this.listenWhen,
  }) : super(key: key, child: child);

  /// The widget which will be rendered as a descendant of the
  /// [IsolateBlocListenerBase].
  final Widget child;

  /// The [isolateBloc] whose `state` will be listened to.
  /// Whenever the [isolateBloc]'s `state` changes, [listener] will be invoked.
  final IsolateBlocWrapper<S> isolateBloc;

  /// The [BlocWidgetListener] which will be called on every `state` change.
  /// This [listener] should be used for any code which needs to execute
  /// in response to a `state` change.
  final BlocWidgetListener<S> listener;

  /// {@macro bloc_listener_listen_when}
  final BlocListenerCondition<S> listenWhen;

  @override
  SingleChildState<IsolateBlocListenerBase<C, S>> createState() =>
      _BlocListenerBaseState<C, S>();
}

class _BlocListenerBaseState<C extends IsolateBloc<Object, S>, S>
    extends SingleChildState<IsolateBlocListenerBase<C, S>> {
  StreamSubscription<S> _subscription;
  S _previousState;
  IsolateBlocWrapper<S> _isolateBlocWrapper;

  @override
  void initState() {
    super.initState();
    _isolateBlocWrapper = widget.isolateBloc ?? context.isolateBloc<C>();
    _previousState = _isolateBlocWrapper?.state;
    _subscribe();
  }

  @override
  void didUpdateWidget(IsolateBlocListenerBase<C, S> oldWidget) {
    super.didUpdateWidget(oldWidget);
    // ignore: close_sinks
    final oldCubit = oldWidget.isolateBloc ?? context.isolateBloc<C>();
    // ignore: close_sinks
    final currentCubit = widget.isolateBloc ?? oldCubit;
    if (oldCubit != currentCubit) {
      if (_subscription != null) {
        _unsubscribe();
        _isolateBlocWrapper = widget.isolateBloc ?? context.isolateBloc<C>();
        _previousState = _isolateBlocWrapper?.state;
      }
      _subscribe();
    }
  }

  @override
  Widget buildWithChild(BuildContext context, Widget child) => child;

  @override
  void dispose() {
    _unsubscribe();
    super.dispose();
  }

  void _subscribe() {
    if (_isolateBlocWrapper != null) {
      _subscription = _isolateBlocWrapper.listen((state) {
        if (widget.listenWhen?.call(_previousState, state) ?? true) {
          widget.listener(context, state);
        }
        _previousState = state;
      });
    }
  }

  void _unsubscribe() {
    if (_subscription != null) {
      _subscription.cancel();
      _subscription = null;
    }
  }
}
